import torch
import numpy as np
import pandas as pd
import networkx as nx
from lib import torch_nvd as nvd
from scipy.stats import fit, beta

def corr(tensor, i, j, W = None):
   numerator = (W * torch.outer(tensor.node_vects[:,i], tensor.node_vects[:,j])).sum()
   denominator_src = torch.sqrt((W * torch.outer(tensor.node_vects[:,i], tensor.node_vects[:,i])).sum())
   denominator_trg = torch.sqrt((W * torch.outer(tensor.node_vects[:,j], tensor.node_vects[:,j])).sum())
   return numerator / (denominator_src * denominator_trg)

def bootstrap(corr, tensor, i, j, W = None):
   better_than_obs = 0
   denominator_src = torch.sqrt((W * torch.outer(tensor.node_vects[:,i], tensor.node_vects[:,i])).sum())
   for _ in range(1000):
      rnd_j = tensor.node_vects[:,j][torch.randperm(tensor.node_vects[:,j].shape[0])]
      numerator = (W * torch.outer(tensor.node_vects[:,i], rnd_j)).sum()
      denominator_trg = torch.sqrt((W * torch.outer(rnd_j, rnd_j)).sum())
      rnd_corr = numerator / (denominator_src * denominator_trg)
      if rnd_corr >= corr:
         better_than_obs += 1
   return better_than_obs / 1000

genres = pd.read_csv("../data/genres.tsv", sep = "\t")#.set_index("id")
regions = pd.read_csv("../data/regions.tsv", sep = "\t")
edges = pd.read_csv("../outputs/projection_band.txt", sep = "\t")
edges = nx.from_pandas_edgelist(edges, source = "src", target = "trg", edge_attr = ["nij", "score"])

nodelist = list(genres["id"].sort_values())
nodemap = {nodelist[i]: i for i in range(len(nodelist))}

genres["id"] = genres["id"].map(nodemap)
regions["id"] = regions["id"].map(nodemap)
genres = genres.sort_values(by = "id").set_index("id")
regions = regions.sort_values(by = "id").set_index("id")
genres = genres.reindex(sorted(genres.columns), axis = 1)
regions = regions.reindex(sorted(regions.columns), axis = 1)
genremap = {i: genres.columns[i] for i in range(len(genres.columns))}
regionmap = {i + len(genres.columns): regions.columns[i] for i in range(len(regions.columns))}

# Normalization not to let this being dominated by prolific bands
genres = (genres.T / genres.sum(axis = 1)).T.fillna(0.0)
# Normalization not to let this being dominated by popular genres
genres = genres / genres.sum(axis = 0)

regions /= regions.sum()
edges = nx.relabel_nodes(edges, nodemap)

tensor = nvd.make_tensor(edges, genres.join(regions))
Linv = nvd._Linv(tensor)
ER = nvd._er(tensor, Linv)
W = 1 / torch.exp(ER)

correlation_matrix = torch.zeros(genres.shape[1], regions.shape[1], device = "cuda")
for i in range(genres.shape[1]):
   for j in range(regions.shape[1]):
      correlation_matrix[i, j] = corr(tensor, i, j + genres.shape[1], W = W)

correlation_matrix = pd.DataFrame(data = correlation_matrix.cpu().numpy(), index = genres.columns, columns = regions.columns)

genre2id = {genremap[genreid]: genreid for genreid in genremap}
region2id = {regionmap[regionid]: regionid for regionid in regionmap}

print("\\begin{table}")
print("\\centering")
print("\\begin{tabular}{l|rl}")
print("Region & Genre & Significance\\\\")
print("\\hline")

while correlation_matrix.shape[1] > 0:
   region = correlation_matrix.max().idxmax()
   genre = correlation_matrix[region].idxmax()
   pvalue = bootstrap(correlation_matrix.loc[genre, region], tensor, genre2id[genre], region2id[region], W = W)
   keep = False
   if pvalue <= 0.01:
      stars = "$^{***}$"
      keep = True
   elif pvalue <= 0.05:
      stars = "$^{**}$"
      keep = True
   elif pvalue <= 0.1:
      stars = "$^{*}$"
      keep = True
   else:
      stars = ""
   if keep:
      print(f"{region.title()} & {genre} & {stars}\\\\")
      correlation_matrix = correlation_matrix.drop(region, axis = 1).drop(genre)
   else:
      correlation_matrix.loc[genre, region] = 0

print("\\end{tabular}")
print("\\caption{The most popular genre with the strongest and significant correlation with a given region. The third column shows the significance level of the correlation, estimated via bootstrapping: $^{*}$p$<$0.1; $^{**}$p$<$0.05; $^{***}$p$<$0.01}")
print("\\label{tab:genre-region-correlations}")
print("\\end{table}")